
import os
import math
import signal
from multiprocessing import Pool
import multiprocessing as mp
from time import time


from modules import help_functions
from modules import cluster


def reads_to_clusters_helper(arguments):
    for k,v in arguments.items():
        args, kwargs = v
    return cluster.reads_to_clusters(*args, **kwargs)


def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def batch_list(lst, nr_cores=1, batch_type = "nr_reads" , merge_consecutive = False ):
    if merge_consecutive:
        batch_id = 2
        batch = []
        for info in lst:
            if info[1] <= batch_id:
                batch.append(info)
            else: # first sequence in new batch
                yield batch
                batch_id += 2
                batch = []
                batch.append(info)
        yield batch

    else:
        if batch_type == "nr_reads":
            l = len(lst)
            chunk_size = int(l/nr_cores) + 1
            for ndx in range(0, l, chunk_size):
                yield lst[ndx:min(ndx + chunk_size, l)]
        
        elif batch_type == "total_nt":
            tot_length = sum([len(seq) for i, b_i, acc, seq, qual, score in lst] )
            nt_chunk_size = int(tot_length/nr_cores) + 1

            batch = []
            curr_size = 0
            for info in lst:
                curr_size += len(info[3])
                batch.append(info)
                if curr_size >= nt_chunk_size:
                    yield batch
                    batch = []
                    curr_size = 0
            yield batch   
        
        elif batch_type == "read_lengths_squared":
            tot_length = sum([ math.pow(len(seq),2) for i, b_i, acc, seq, qual, score in lst] )
            nt_chunk_size = int(tot_length/nr_cores) + 1
            batch = []
            curr_size = 0
            for info in lst:
                curr_size +=  math.pow(len(info[3]),2) 
                batch.append(info)
                if curr_size >= nt_chunk_size:
                    yield batch
                    batch = []
                    curr_size = 0
            yield batch               



def print_intermediate_results(clusters, cluster_seq_origin, args, iter_nr):
    path = args.outfolder +"/{0}".format(iter_nr) 
    help_functions.mkdir_p( path )
    outfile = open(os.path.join(path, "pre_clusters.csv"), "w")
    nontrivial_cluster_index = 0
    for c_id, all_read_acc in sorted(clusters.items(), key = lambda x: len(x[1]), reverse=True):
        for r_acc in all_read_acc:
            outfile.write("{0}\t{1}\n".format(c_id, "_".join([item for item in r_acc.split("_")[:-1]]) ))
        if len(all_read_acc) > 1:
            nontrivial_cluster_index += 1
    print("Nr clusters larger than 1:", nontrivial_cluster_index) #, "Non-clustered reads:", len(archived_reads))
    print("Nr clusters (all):", len(clusters)) #, "Non-clustered reads:", len(archived_reads))


    origins_outfile = open(os.path.join(path,  "cluster_origins.csv"), "w")
    for cl_id, all_read_acc in sorted(clusters.items(), key = lambda x: len(x[1]), reverse=True):
        read_cl_id, b_i, acc, c_seq, c_qual, score, error_rate = cluster_seq_origin[cl_id]
        origins_outfile.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(read_cl_id, acc, c_seq, c_qual, score, error_rate)) 
    outfile.close()
    origins_outfile.close()


def parallel_clustering(read_array, p_emp_probs, args):
    num_batches = args.nr_cores 
    read_batches = [batch for batch in batch_list(read_array, num_batches, batch_type = args.batch_type )]
    print("Using total nucleotide batch sizes:", [sum([len(seq) for i, b_i, acc, seq, qual, score in b]) for b in read_batches] )
    print("Nr reads in batches:", [len(b)  for b in read_batches] )
    cluster_batches = []
    cluster_seq_origin_batches = []
    lowest_batch_index_db = []
    for batch in read_batches:
        tmp_clust = {}
        tmp_clust_origin = {}
        for i, b_i, acc, seq, qual, score in batch:
            tmp_clust[i] = [acc]
            tmp_clust_origin[i] = (i, b_i, acc, seq, qual, score)
        cluster_batches.append(tmp_clust)
        cluster_seq_origin_batches.append(tmp_clust_origin)
        lowest_batch_index_db.append({})
    del read_array

    ####### parallelize alignment #########
    # pool = Pool(processes=mp.cpu_count())
    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGINT, original_sigint_handler)
    try:
        mp.set_start_method('spawn')
        print("Environment set:", mp.get_context())
    except RuntimeError:
        print("Environment already set:", mp.get_context())
    it = 1
    while True:
        # Structure up batches
        print()
        print("ITERATION", it)
        print("Using {0} batches.".format(num_batches))

        if len(read_batches) == 1:
            start_cluster = time()

            data = {i+1 :((cluster_batches[0], cluster_seq_origin_batches[0], read_batches[0], p_emp_probs, lowest_batch_index_db[0], 1, args), {})} 
            result = reads_to_clusters_helper(data) # { new_batch_index : (Cluster, cluster_seq_origin, H, new_batch_index)}
            Cluster, cluster_seq_origin, _, _ = result[1]
            print("Time elapesd clustering last iteration single core:", time() - start_cluster)
            return Cluster, cluster_seq_origin


        start_multi = time()
        pool = Pool(processes=int(num_batches))
        try:
            # print([len(b) for b in read_batches])
            data = [ {i+1 :((cluster_batches[i], cluster_seq_origin_batches[i], read_batches[i], p_emp_probs, lowest_batch_index_db[i], i+1, args), {})} for i in range(len(read_batches))]
            res = pool.map_async(reads_to_clusters_helper, data)
            cluster_results =res.get(999999999) # Without the timeout this blocking call ignores all signals.
        except KeyboardInterrupt:
            print("Caught KeyboardInterrupt, terminating workers")
            pool.terminate()
            sys.exit()
        else:
            pool.close()
        pool.join()

        print("Time elapesd multiprocessing:", time() - start_multi)

        start_joining = time()
        all_repr = [] # all_repr = [top_new_seq_origins]
        all_cl = []
        all_minimizer_databases = {}
        for output_dict in cluster_results:
            print("New batch")
            for k, v in output_dict.items():
                new_clusters, new_representatives, minimizer_database_new, batch_index = v
                print("Batch index", k)
            # for new_clusters, new_representatives, minimizer_database_new, batch_index in cluster_results: 
                all_cl.append(new_clusters)
                all_repr.append(new_representatives)
                all_minimizer_databases[batch_index] =  minimizer_database_new

        all_clusters = merge_dicts(*all_cl)
        all_representatives = merge_dicts(*all_repr)
        read_array =  [ (i, b_index, acc, seq, qual, score) for i, (i, b_index, acc, seq, qual, score, error_rate) in sorted(all_representatives.items(), key=lambda x: x[1][5], reverse=True)] 
        new_nr_repr = len(read_array)
        print("number of representatives left to cluster:", new_nr_repr)
        print("Time elapesd joining clusters:", time() - start_joining)

        # Determine new number of batches
        if num_batches == 1:
            return all_clusters, all_representatives
        else:
            print_intermediate_results(all_clusters, all_representatives, args, it)

        it += 1
        read_batches = [batch for batch in batch_list(read_array, num_batches, batch_type = args.batch_type, merge_consecutive = True)]
        num_batches = len(read_batches)
        print("Batches after pairwise consecutive merge:", num_batches)
        print("Using total nucleotide batch sizes:", [sum([len(seq) for i, b_i, acc, seq, qual, score in b]) for b in read_batches] )
        print("Using nr reads batch sizes:", [len(b)  for b in read_batches] )
        cluster_batches = []
        cluster_seq_origin_batches = []
        lowest_batch_index_db = []
        for batch in read_batches:
            tmp_clust = {}
            tmp_clust_origin = {}
            lowest_batch_index = min( [ prev_batch_index for (read_cl_id, prev_batch_index, acc, seq, qual, score) in batch ] )
            for i, b_i, acc, seq, qual, score in batch:
                tmp_clust[i] = all_clusters[i]
                tmp_clust_origin[i] = all_representatives[i]
            cluster_batches.append(tmp_clust)
            cluster_seq_origin_batches.append(tmp_clust_origin)
            lowest_batch_index_db.append( all_minimizer_databases[lowest_batch_index])

        del all_minimizer_databases




from __future__ import print_function
import subprocess
import sys, os
from sys import stdout
import re 
import shutil
import parasail
import glob

from modules import help_functions


def cigar_to_seq(cigar, query, ref):
    cigar_tuples = []
    result = re.split(r'[=DXSMI]+', cigar)
    cig_pos = 0
    for length in result[:-1]:
        cig_pos += len(length)
        type_ = cigar[cig_pos]
        cig_pos += 1
        cigar_tuples.append((int(length), type_ ))

    r_index = 0
    q_index = 0
    q_aln = []
    r_aln = []
    for length_ , type_ in cigar_tuples:
        if type_ == "=" or type_ == "X":
            q_aln.append(query[q_index : q_index + length_])
            r_aln.append(ref[r_index : r_index + length_])

            r_index += length_
            q_index += length_
        
        elif  type_ == "I":
            # insertion w.r.t. reference
            r_aln.append('-' * length_)
            q_aln.append(query[q_index: q_index + length_])
            #  only query index change
            q_index += length_

        elif type_ == 'D':
            # deletion w.r.t. reference
            r_aln.append(ref[r_index: r_index + length_])
            q_aln.append('-' * length_)
            #  only ref index change
            r_index += length_
        
        else:
            print("error")
            print(cigar)
            sys.exit()

    return  "".join([s for s in q_aln]), "".join([s for s in r_aln]), cigar_tuples


def parasail_alignment(s1, s2, match_score = 2, mismatch_penalty = -2, opening_penalty = 3, gap_ext = 1):
    user_matrix = parasail.matrix_create("ACGT", match_score, mismatch_penalty)
    result = parasail.sg_trace_scan_16(s1, s2, opening_penalty, gap_ext, user_matrix)
    if result.saturated:
        print("SATURATED!",len(s1), len(s2))
        result = parasail.sg_trace_scan_32(s1, s2, opening_penalty, gap_ext, user_matrix)
        print("computed 32 bit instead")

    # difference in how to obtain string from parasail between python v2 and v3... 
    if sys.version_info[0] < 3:
        cigar_string = str(result.cigar.decode).decode('utf-8')
    else:
        cigar_string = str(result.cigar.decode, 'utf-8')
    s1_alignment, s2_alignment, cigar_tuples = cigar_to_seq(cigar_string, s1, s2)
    # print(result.score, len(s1), len(s2))
    # print(s1_alignment)
    # print(s2_alignment)
    # print(cigar_string)
    # sys.exit()

    return s1_alignment, s2_alignment, cigar_string, cigar_tuples, result.score

def reverse_complement(string):
    #rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N':'N', 'X':'X'}
    # Modified for Abyss output
    rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'a':'t', 'c':'g', 'g':'c', 't':'a', 'N':'N', 'X':'X', 'n':'n', 'Y':'R', 'R':'Y', 'K':'M', 'M':'K', 'S':'S', 'W':'W', 'B':'V', 'V':'B', 'H':'D', 'D':'H', 'y':'r', 'r':'y', 'k':'m', 'm':'k', 's':'s', 'w':'w', 'b':'v', 'v':'b', 'h':'d', 'd':'h'}

    rev_comp = ''.join([rev_nuc[nucl] for nucl in reversed(string)])
    return(rev_comp)

def run_spoa(reads, spoa_out_file, spoa_path):
    with open(spoa_out_file, "w") as output_file:
        # print('Running spoa...', end=' ')
        stdout.flush()
        null = open("/dev/null", "w")
        subprocess.check_call([ spoa_path, reads, "-l", "0", "-r", "0", "-g", "-2"], stdout=output_file, stderr=null)
        # print('Done.')
        stdout.flush()
    output_file.close()
    l = open(spoa_out_file, "r").readlines()
    consensus = l[1].strip()
    return consensus

def run_medaka(reads_to_center, center_file, outfolder, cores, medaka_model):
    medaka_stdout = os.path.join(outfolder, "stdout.txt")
    with open(medaka_stdout, "w") as output_file:
        # print('Running medaka...', end=' ')
        stdout.flush()
        medaka_stderr = open(os.path.join(outfolder, "stderr.txt"), "w")
        if medaka_model:
            subprocess.check_call(['medaka_consensus', '-i', reads_to_center, "-d", center_file, "-o", outfolder, "-t", cores, "-m", medaka_model], stdout=output_file, stderr=medaka_stderr)
        else:
            subprocess.check_call(['medaka_consensus', '-i', reads_to_center, "-d", center_file, "-o", outfolder, "-t", cores], stdout=output_file, stderr=medaka_stderr)

        # print('Done.')
        stdout.flush()
    output_file.close()

def run_racon(reads_to_center, center_file, outfolder, cores, racon_iter):
    racon_stdout = os.path.join(outfolder, "stdout.txt")
    with open(racon_stdout, "w") as output_file:
        # print('Running medaka...', end=' ')
        stdout.flush()
        for i in range(racon_iter):
            read_alignments = open(os.path.join(outfolder, "read_alignments_it_{0}.paf".format(i)), 'w')
            mm2_stderr = open(os.path.join(outfolder, "mm2_stderr_it_{0}.txt".format(i)), "w")
            racon_stderr = open(os.path.join(outfolder, "racon_stderr_it_{0}.txt".format(i)), "w")
            racon_polished = open(os.path.join(outfolder, "racon_polished_it_{0}.fasta".format(i)), 'w')

            subprocess.check_call(['minimap2', '-x', 'map-ont', center_file, reads_to_center], stdout=read_alignments, stderr=mm2_stderr)
            subprocess.check_call(['racon', reads_to_center, read_alignments.name, center_file], stdout=racon_polished, stderr=racon_stderr)
            center_file = racon_polished.name

        shutil.copyfile(center_file, os.path.join(outfolder, "consensus.fasta"))
        # print('Done.')
        stdout.flush()
    output_file.close()
    # consensus.run_medaka( " medaka_consensus -i ~/tmp/stefan_isonclust/mixed_b1_b3_b4_b5.fastq -d ~/tmp/stefan_isonclust/mixed_b1_b3_b4_b5_isonclust/consensus_references.fasta -o ~/tmp/stefan_isonclust/mixed_b1_b3_b4_b5_medaka/ -t 1 -m r941_min_high_g303")


def detect_reverse_complements(centers, rc_identity_threshold):
    filtered_centers = []
    already_removed = set()
    for i, (nr_reads_in_cl, c_id, seq, reads_path) in enumerate(centers):
        if type(reads_path) != list:
            all_reads = [reads_path]
        else:
            all_reads = reads_path

        merged_cluster_id = c_id
        merged_nr_reads = nr_reads_in_cl
        if c_id in already_removed:
            print("has already been merged, skipping")
            continue

        elif i == len(centers) - 1: # last sequence and it is not in already removed
            filtered_centers.append( [merged_nr_reads, c_id, seq, all_reads ] )

        else:
            for j, (nr_reads_in_cl2, c_id2, seq2, reads_path) in enumerate(centers[i+1:]):
                seq2_rc = reverse_complement(seq2)
                seq_aln, seq2_rc_aln, cigar_string, cigar_tuples, alignment_score = parasail_alignment(seq, seq2_rc)
                # print(i,j)
                # print(seq_aln)
                # print(seq2_rc_aln)
                nr_mismatching_pos = len([1 for n1, n2 in zip(seq_aln, seq2_rc_aln) if n1 != n2])
                total_pos = len(seq_aln)
                aln_identity =  (total_pos - nr_mismatching_pos) / float(total_pos)
                print(aln_identity)

                if aln_identity >= rc_identity_threshold:
                    print("Detected alignment identidy above threchold for reverse complement. Keeping center with the most read support and adding rc reads to supporting reads.")
                    merged_nr_reads += nr_reads_in_cl2
                    already_removed.add(c_id2)

                    if type(reads_path) != list:
                        all_reads.append(reads_path)
                    else:
                        for rp in reads_path:
                            all_reads.append(rp)

            filtered_centers.append( [merged_nr_reads, c_id, seq, all_reads] )

    print(len(filtered_centers), "consensus formed.")
    return filtered_centers


def polish_sequences(centers, args):
    print("Saving spoa references to files:", os.path.join(args.outfolder, "consensus_reference_X.fasta"))
    # printing output from spoa and grouping reads
    # to_polishing = []
    if args.medaka:
        polishing_pattern = os.path.join(args.outfolder, "medaka_cl_id_*")
    elif args.racon:
        polishing_pattern = os.path.join(args.outfolder, "racon_cl_id_*")

    for folder in glob.glob(polishing_pattern):
        shutil.rmtree(folder)

    spoa_pattern = os.path.join(args.outfolder, "consensus_reference_*")
    for file in glob.glob(spoa_pattern):
        os.remove(file)

    for i, (nr_reads_in_cluster, c_id, center, all_reads) in enumerate(centers):
        # print('lol',c_id,center)
        spoa_center_file = os.path.join(args.outfolder, "consensus_reference_{0}.fasta".format(c_id))
        f = open(spoa_center_file, "w")
        f.write(">{0}\n{1}\n".format("consensus_cl_id_{0}_total_supporting_reads_{1}".format(c_id, nr_reads_in_cluster), center))
        f.close()
        
        all_reads_file = os.path.join(args.outfolder, "reads_to_consensus_{0}.fastq".format(c_id))
        f = open(all_reads_file, "w")
        for fasta_file in all_reads: 
            reads = { acc : (seq, qual) for acc, (seq, qual) in help_functions.readfq(open(fasta_file, 'r'))}
            for acc, (seq, qual) in reads.items():
                acc_tmp = acc.split()[0]
                f.write("@{0}\n{1}\n{2}\n{3}\n".format(acc_tmp, seq, "+", qual))
        f.close()
        # to_polishing.append( (nr_reads_in_cluster, c_id, spoa_center_file, all_reads_file) )

        if args.medaka:
            print("running medaka on spoa reference {0}.".format(c_id))
            # for (nr_reads_in_cluster, c_id, spoa_center_file, all_reads_file) in to_polishing:
            polishing_outfolder = os.path.join(args.outfolder, "medaka_cl_id_{0}".format(c_id))
            help_functions.mkdir_p(polishing_outfolder)
            run_medaka(all_reads_file, spoa_center_file, polishing_outfolder, "1", args.medaka_model)
            print("Saving medaka reference to file:", os.path.join(args.outfolder, "medaka_cl_id_{0}/consensus.fasta".format(c_id)))   
            l = open(os.path.join(polishing_outfolder, "consensus.fasta"), 'r').readlines()
            center_polished = l[1].strip()
            centers[i][2] = center_polished
        elif args.racon:
            print("running racon on spoa reference {0}.".format(c_id))
            # for (nr_reads_in_cluster, c_id, spoa_center_file, all_reads_file) in to_polishing:
            polishing_outfolder = os.path.join(args.outfolder, "racon_cl_id_{0}".format(c_id))
            help_functions.mkdir_p(polishing_outfolder)
            run_racon(all_reads_file, spoa_center_file, polishing_outfolder, "1", args.racon_iter)
            print("Saving racon reference to file:", os.path.join(args.outfolder, "racon_cl_id_{0}/consensus.fasta".format(c_id)))   
            l = open(os.path.join(polishing_outfolder, "consensus.fasta"), 'r').readlines()
            center_polished = l[1].strip()
            centers[i][2] = center_polished

    f.close()
    return centers


def form_draft_consensus(clusters, representatives, sorted_reads_fastq_file, work_dir, abundance_cutoff, args):
    centers = []
    reads = { acc : (seq, qual) for acc, (seq, qual) in help_functions.readfq(open(sorted_reads_fastq_file, 'r'))}
    for c_id, all_read_acc in sorted(clusters.items(), key = lambda x: (len(x[1]),representatives[x[0]][5]), reverse=True):
        reads_path = open(os.path.join(work_dir, "reads_c_id_{0}.fq".format(c_id)), "w")
        nr_reads_in_cluster = len(all_read_acc)
        # print("nr_reads_in_cluster", nr_reads_in_cluster)
        if nr_reads_in_cluster >= abundance_cutoff:
            for acc in all_read_acc:
                seq, qual = reads[acc]
                reads_path.write("@{0}\n{1}\n{2}\n{3}\n".format(acc, seq, "+", qual))
                # reads_path.write(">{0}\n{1}\n".format(str(q_id)+str(pos1)+str(pos2), seq))
            reads_path.close()
            # spoa_ref = create_augmented_reference.run_spoa(reads_path.name, os.path.join(work_dir,"spoa_tmp.fa"), "spoa")
            print("creating center of {0} sequences.".format(nr_reads_in_cluster))
            center = run_spoa(reads_path.name, os.path.join(work_dir,"spoa_tmp.fa"), "spoa")
            centers.append( [nr_reads_in_cluster, c_id, center, reads_path.name])
    return centers



from __future__ import print_function
from functools import reduce
import os,sys
import argparse
from collections import defaultdict
import math
from collections import deque
import itertools
from operator import mul
import re

import parasail

from modules import help_functions


def get_kmer_minimizers(seq, k_size, w_size):
    # kmers = [seq[i:i+k_size] for i in range(len(seq)-k_size) ]
    w = w_size - k_size
    window_kmers = deque([seq[i:i+k_size] for i in range(w +1)])
    curr_min = min(window_kmers)
    minimizers = [ (curr_min, list(window_kmers).index(curr_min)) ]

    for i in range(w+1,len(seq) - k_size + 1):
        new_kmer = seq[i:i+k_size]
        # updateing window
        discarded_kmer = window_kmers.popleft()
        window_kmers.append(new_kmer)

        # we have discarded previous windows minimizer, look for new minimizer brute force
        if curr_min == discarded_kmer: 
            curr_min = min(window_kmers)
            minimizers.append( (curr_min, list(window_kmers).index(curr_min) + i - w ) )

        # Previous minimizer still in window, we only need to compare with the recently added kmer 
        elif new_kmer < curr_min:
            curr_min = new_kmer
            minimizers.append( (curr_min, i) )

    return minimizers



def get_all_hits(minimizers, minimizer_database, read_cl_id):
    """
        Get all representatives ID's that shares matches with the minimizers in the read.  
    """
    hit_clusters_ids = defaultdict(int)
    hit_clusters_hit_index = defaultdict(list)
    hit_clusters_hit_positions = defaultdict(list)
    for i, (m, pos) in enumerate(minimizers): # iterating over minimizers from upstream to downstream in read
        if m in minimizer_database:
            for cl_id in minimizer_database[m]: 
                hit_clusters_ids[cl_id] += 1
                hit_clusters_hit_index[cl_id].append(i) # index of the minimizer among the coordinate sorted minimizers in the read
                hit_clusters_hit_positions[cl_id].append(pos) # positions of the minimizer among the coordinate sorted minimizers in the read

    if read_cl_id in hit_clusters_ids:
        del hit_clusters_ids[read_cl_id]
        del hit_clusters_hit_index[read_cl_id]
        del hit_clusters_hit_positions[read_cl_id]

    return hit_clusters_ids, hit_clusters_hit_index, hit_clusters_hit_positions




def get_best_cluster(read_cl_id, compressed_seq_len, hit_clusters_ids, hit_clusters_hit_positions, minimizers, nummber_of_minimizers, hit_clusters_hit_index, representatives, p_emp_probs, args):
    """
        Tally up total covered (mapped region) and compare it with total unmapped region. What is counted as a consecutive mapped block depends on minimizer qualitity
        threshold as mapped-read-legnth/total-read-length is chosen to classify a read as belonging to the cluster
        
        Return: An integer >= 0 denoting the cluster ID that this read was assigned to. In not assigend to any previous cluster, return -1. 
                [Also returing mapped ratio and nr shared minimizers to best read for logging purposes.]
    """
    best_cluster_id = -1
    nr_shared_kmers = 0
    mapped_ratio = 0.0
    if hit_clusters_ids:
        top_matches = sorted(hit_clusters_hit_positions.items(), key=lambda x: (len(x[1]), sum(x[1]), representatives[x[0]][2]),  reverse=True) #sorted(hit_clusters_ids.items(), key=lambda x: x[1],  reverse=True)
        top_hits = len(top_matches[0][1])
        nr_shared_kmers = top_hits
        if top_hits < args.min_shared:
            pass
        else:
            for tm in top_matches:
                cl_id = tm[0]
                nm_hits = len(tm[1])
                if nm_hits < args.min_fraction * top_hits or nm_hits < args.min_shared:
                    break

                # cl_size = len(hit_clusters_ids)
                minimizer_hit_positions = hit_clusters_hit_positions[cl_id]
                minimizer_hit_indices = hit_clusters_hit_index[cl_id]
                assert len(minimizer_hit_indices) == len(minimizer_hit_positions)
                _, _, _, _, _, _, error_rate_c = representatives[cl_id]
                _, _, _, _, _, _, error_rate_read = representatives[read_cl_id]
                p_error_in_kmers_emp =  1.0 - p_shared_minimizer_empirical(error_rate_read, error_rate_c, p_emp_probs)
                minimizer_error_probabilities = [p_error_in_kmers_emp]*nummber_of_minimizers
                total_mapped = 0
                # prev_mpos = 0
                prob_all_errors_since_last_hit = [reduce(mul, minimizer_error_probabilities[: minimizer_hit_indices[0]], 1)] +  [ reduce(mul, minimizer_error_probabilities[hit_idx1+1: hit_idx2], 1) for hit_idx1, hit_idx2 in zip(minimizer_hit_indices[:-1], minimizer_hit_indices[1:]) ] + [reduce(mul, minimizer_error_probabilities[minimizer_hit_indices[-1]+1 : ], 1)]

                assert len(prob_all_errors_since_last_hit) == len(minimizer_hit_positions) + 1
                for i in range(len(minimizer_hit_indices)):
                    if prob_all_errors_since_last_hit[i] < args.min_prob_no_hits:
                        pass
                    else:
                        if i == 0:
                            total_mapped += minimizer_hit_positions[i]
                        else:
                            total_mapped += minimizer_hit_positions[i] - minimizer_hit_positions[i-1]
                if prob_all_errors_since_last_hit[-1] < args.min_prob_no_hits:
                    pass
                else:
                    total_mapped += compressed_seq_len - minimizer_hit_positions[-1]

                mapped_ratio = total_mapped /float(compressed_seq_len) 
                
                if mapped_ratio > args.mapped_threshold:
                    # is_covered = True
                    return cl_id, nm_hits, mapped_ratio

    return  best_cluster_id, nr_shared_kmers, mapped_ratio 


def parasail_block_alignment(s1, s2, k, match_id, match_score = 2, mismatch_penalty = -2, opening_penalty = 5, gap_ext = 1):
    user_matrix = parasail.matrix_create("ACGT", match_score, mismatch_penalty)
    result = parasail.sg_trace_scan_16(s1, s2, opening_penalty, gap_ext, user_matrix)
    if result.saturated:
        print("SATURATED!",len(s1), len(s2))
        result = parasail.sg_trace_scan_32(s1, s2, opening_penalty, gap_ext, user_matrix)
        print("computed 32 bit instead")

    # difference in how to obtain string from parasail between python v2 and v3... 
    if sys.version_info[0] < 3:
        cigar_string = str(result.cigar.decode).decode('utf-8')
    else:
        cigar_string = str(result.cigar.decode, 'utf-8')
    
    s1_alignment, s2_alignment = help_functions.cigar_to_seq(cigar_string, s1, s2)

    # Rolling window of matching blocks
    match_vector = [ 1 if n1 == n2 else 0 for n1, n2 in zip(s1_alignment, s2_alignment) ]    
    match_window = deque(match_vector[:k]) # initialization
    current_match_count = sum(match_window)
    aligned_region = []
    if current_match_count >= match_id:
        aligned_region.append(1)
    else:
        aligned_region.append(0)


    for new_m_state in match_vector[k:]:
        prev_m_state = match_window.popleft()
        current_match_count = current_match_count - prev_m_state + new_m_state 
        match_window.append(new_m_state)
        
        if current_match_count >= match_id:
            aligned_region.append(1)
        else:        
            aligned_region.append(0)

    # print("".join([str(m) for m in aligned_region]))
    # print("Aligned ratio (tot aligned/len(seq1):", sum(aligned_region)/float(len(s1)))
    alignment_ratio = sum(aligned_region)/float(len(s1))
    return (s1, s2, (s1_alignment, s2_alignment, alignment_ratio))


def get_best_cluster_block_align(read_cl_id, representatives, hit_clusters_ids, hit_clusters_hit_positions, phred_char_to_p, args):
    best_cluster_id = -1
    top_matches = sorted(hit_clusters_hit_positions.items(), key=lambda x: (len(x[1]), sum(x[1]), representatives[x[0]][2]),  reverse=True) #sorted(hit_clusters_ids.items(), key=lambda x: x[1],  reverse=True)
    _, _, _, seq, r_qual, _, _ = representatives[read_cl_id]
    # print(top_matches)
    top_hits = len(top_matches[0][1])
    alignment_ratio = 0.0
    for tm in top_matches:
        cl_id = tm[0]
        nm_hits = len(tm[1])
        if nm_hits < top_hits:
            break
        _, _, _, c_seq, c_qual, _, _ = representatives[cl_id]

        poisson_mean = sum([ r_qual.count(char_) * phred_char_to_p[char_] for char_ in set(r_qual)])
        poisson_mean2 = sum([ c_qual.count(char_) * phred_char_to_p[char_] for char_ in set(c_qual)])

        error_rate_sum = poisson_mean/float(len(seq)) + poisson_mean2/float(len(c_seq))  # k = max(int(mean_plus_two_stdvs_q2 + mean_plus_two_stdvs_q1) + 1 + int(len(seq)*args.variant_rate) , 40)
        if error_rate_sum <= 0.01:
            gap_opening_penalty = 5
        elif  0.01 < error_rate_sum <= 0.04:
            gap_opening_penalty = 4
        elif  0.04 < error_rate_sum <= 0.1:
            gap_opening_penalty = 3
        elif  0.1 < error_rate_sum:
            gap_opening_penalty = 2

        match_id_tailored = math.floor((1.0 - error_rate_sum) * args.k)
        (s1, s2, (s1_alignment, s2_alignment, alignment_ratio)) = parasail_block_alignment(seq, c_seq, args.k, match_id_tailored, opening_penalty = gap_opening_penalty,  )
        # print("Expected errors:", poisson_mean, poisson_mean2)
        if alignment_ratio >= args.aligned_threshold: #args.mapped_threshold:
            return cl_id, nm_hits,  error_rate_sum, s1_alignment, s2_alignment, alignment_ratio

    return  best_cluster_id, 0,  -1, -1, -1, alignment_ratio

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def reads_to_clusters(clusters, representatives, sorted_reads, p_emp_probs, minimizer_database, new_batch_index, args):
    """
        Iterates throughreads in sorted order (w.r.t. score) and:
            1. homopolymenr compresses the read and obtain minimizers
            2. Finds the homopolymenr compressed error rate (if not computed in previous pass if more than 1 core specified to the program)
            3. Finds all the representatives that shares minimizers with the read
            4. Finds the best of the hits using mapping approach
            5. If no hit is found in 4. tries to align to representative with th most shared minimizers.
            6. Adds current read to representative, or makes it a new representative of a new cluster.
            7. If new representative: add the minimizers to the minimizer database
            8. Assign the actual reads to their new cluster and their new cluster representative (since all reads were initialized as their own representatives to deal with multiprocessing) 
    """

    ## For multiprocessing only
    prev_b_indices = [ prev_batch_index for (read_cl_id, prev_batch_index, acc, seq, qual, score) in sorted_reads ]
    lowest_batch_index = max(1, min(prev_b_indices))
    skip_count = prev_b_indices.count(lowest_batch_index)
    print("Saved: {0} iterations.".format(skip_count) )
    ###################################
    
    ## logging counters 
    aln_passed_criteria = 0
    mapped_passed_criteria = 0
    aln_called = 0
    ###################
    
    phred_char_to_p = {chr(i) : min( 10**( - (ord(chr(i)) - 33)/10.0 ), 0.79433)  for i in range(128)} # PHRED encoded quality character to prob of error. Need this locally if multiprocessing
    cluster_to_new_cluster_id = {}

    if args.print_output:
        eprint("Iteration\tNrClusters\tMinDbSize\tCurrReadId\tClusterSizes")

    for i, (read_cl_id, prev_batch_index, acc, seq, qual, score) in enumerate(sorted_reads):

        ## This if statement is only active in parallelization code 
        ## to keep track of already processed reads in previous iteration
        if prev_batch_index == lowest_batch_index:
            lst = list(representatives[read_cl_id])
            lst[1] = new_batch_index
            t = tuple(lst)
            representatives[read_cl_id] =  t # just updated batch index
            continue
        ##############################################################
        
        ################################################################################
        ############  Just for develop purposes, print some info to std out ############
        if i % args.print_output == 0: 
            inv_map = {}
            for k, v in cluster_to_new_cluster_id.items():
                inv_map.setdefault(v, set()).add(k)
            cl_tmp = sorted( [ 1 + sum([len(clusters[cl_id]) for cl_id in c ]) for c in inv_map.values() ], reverse= True)
            cl_tmp_nontrivial = [cl_size_tmp for cl_size_tmp in cl_tmp if cl_size_tmp > 1]
            eprint("{0}\t{1}\t{2}\t{3}\t{4}".format(i, len(cl_tmp_nontrivial), len(minimizer_database), "_".join(acc.split("_")[:-1]), ",".join([str(s_) for s_ in sorted(cl_tmp_nontrivial, reverse=True)])))
            # print("Processing read", i+1 , "seq length:", len(seq), "nr non-trivial clusters:", len(cl_tmp_nontrivial), "kmers stored:", len(minimizer_database))
            # print("Non trivial cluster sizes:", sorted(cl_tmp_nontrivial, reverse=True))

            # print("clust distr:", [c_len for c_len in cl_tmp if c_len > 100] )
            # depth = [len(nr_cl) for kmer, nr_cl in  sorted(minimizer_database.items(), key=lambda x: len(x[1]), reverse= True) if len(nr_cl) > 1 ]
            # print("Minimizer database depths:", depth)
            # print("Nr trivial clusters :", cl_sizes)
        ################################################################################
        ################################################################################

        # 1. homopolymenr compress read and obtain minimizers

        seq_hpol_comp = ''.join(ch for ch, _ in itertools.groupby(seq))
        if len(seq_hpol_comp) < args.k:
            print( "skipping read of length:", len(seq), "homopolymer compressed:", len(seq_hpol_comp), seq)
            continue 
        minimizers = get_kmer_minimizers(seq_hpol_comp, args.k, args.w)
        
        # 2. Find the homopolymer compressed error rate (else statement is the only one active in single core mode)

        if len(representatives[read_cl_id]) == 7: # we have already computed homopolymenr compressed error rate in previous iteration (if isONclust is called with multiple cores):
            lst = list(representatives[read_cl_id])
            lst[1] = new_batch_index
            t = tuple(lst)
            representatives[read_cl_id] =  t # just updated batch index
        else:
            # indices = [i for i, (n1,n2) in enumerate(zip(seq[:-1],seq[1:])) if n1 != n2] # indicies we want to take quality values from to get quality string of homopolymer compressed read 
            # indices.append(len(seq) - 1)
            # qualcomp = ''.join([qual[i] for i in indices])
            # assert len(seq_hpol_comp) == len(qualcomp)
            all_read_hpol_lengths = [len([c for c in g]) for ch, g in itertools.groupby(seq)]
            # print(all_read_hpol_lengths)
            qualcomp = []
            start = 0
            for h_len in all_read_hpol_lengths:
                q_max = min(qual[start: start + h_len], key = lambda x: phred_char_to_p[x])
                qualcomp.append(q_max)
                # if h_len > 2:
                #     print(qual[start: start + h_len], q_max)
                start += h_len
            qualcomp = "".join([q for q in qualcomp])
            assert len(seq_hpol_comp) == len(qualcomp)
            # print(qualcomp)
            # assert len(qualcomp) == len(qualcomp2)
            # print(qualcomp)
            # print(qualcomp2)
            # print()

            # compute the average error rate after compression
            poisson_mean = sum([ qualcomp.count(char_) * phred_char_to_p[char_] for char_ in set(qualcomp)])
            h_pol_compr_error_rate = poisson_mean/float(len(qualcomp))
            representatives[read_cl_id] = (read_cl_id, new_batch_index, acc, seq, qual, score, h_pol_compr_error_rate) # adding homopolymenr compressed error rate to info tuple of cluster origin sequence


        # 3. Find all the representatives with shared minimizers (this is the time consuming function for noisy and large datasets)

        hit_clusters_ids, hit_clusters_hit_index, hit_clusters_hit_positions = get_all_hits(minimizers, minimizer_database, read_cl_id)

        
        # 4. Finds the best of the hits using mapping approach

        best_cluster_id_m, nr_shared_kmers_m, mapped_ratio = get_best_cluster(read_cl_id, len(seq_hpol_comp), hit_clusters_ids, hit_clusters_hit_positions, minimizers, len(minimizers), hit_clusters_hit_index, representatives, p_emp_probs, args)
        
        
        # 5. If step 4 is unsuccessfull we try to align the read to the representative(s) with the most shared minimizers.

        if best_cluster_id_m >= 0:
            mapped_passed_criteria += 1

        if best_cluster_id_m < 0 and nr_shared_kmers_m >= args.min_shared:
            aln_called += 1
            best_cluster_id_a, nr_shared_kmers_a, error_rate_sum, s1_alignment, s2_alignment, alignment_ratio = get_best_cluster_block_align(read_cl_id, representatives, hit_clusters_ids, hit_clusters_hit_positions, phred_char_to_p, args)
            if best_cluster_id_a >= 0:
                aln_passed_criteria += 1

        else:
            best_cluster_id_a = -1


        # 6. Adds current read to representative, or makes it a new representative of a new cluster.
        
        best_cluster_id = max(best_cluster_id_m,best_cluster_id_a)
        if best_cluster_id >= 0:
            cluster_to_new_cluster_id[read_cl_id] = best_cluster_id

        # 7. If new representative: add the minimizers to the minimizer database. 

        else :  # Stays in current cluser, adding representative minimixers
            for m, pos in minimizers:
                if m in minimizer_database:
                    minimizer_database[m].add(read_cl_id)
                else:
                    minimizer_database[m] = set()
                    minimizer_database[m].add(read_cl_id)     
    
    
    # 8. Since all reads were initialized as their own representatives we need to reassign reads to their new representative, (this approach was implemented to deal with iterative assigment in the multiprocessing version)
    for read_cl_id in cluster_to_new_cluster_id:
        new_cl_id = cluster_to_new_cluster_id[read_cl_id]
        all_reads = clusters[read_cl_id]
        for read_acc in all_reads:
            clusters[new_cl_id].append(read_acc)
        del clusters[read_cl_id]
        # delete old origins
        del representatives[read_cl_id]
    ##########################

    print("Total number of reads iterated through:{0}".format(len(sorted_reads)))
    print("Passed mapping criteria:{0}".format(mapped_passed_criteria))
    print("Passed alignment criteria in this process:{0}".format(aln_passed_criteria))
    print("Total calls to alignment mudule in this process:{0}".format(aln_called))

    return { new_batch_index : (clusters, representatives, minimizer_database, new_batch_index)}


def p_shared_minimizer_empirical(error_rate_read, error_rate_center, p_emp_probs):
    e1 = round(error_rate_read, 2)
    # print(e1)
    if e1 > 0.15:
        e1 = 0.15
    if e1 < 0.01:
        e1 = 0.01
    e2 = round(error_rate_center, 2)
    # print(e2)
    if e2 > 0.15:
        e2 = 0.15
    if e2 < 0.01:
        e2 = 0.01
    # print(e2)
    p_kmer_shared = p_emp_probs[(e1,e2)]
    return p_kmer_shared








import os
import errno
import re



'''
    Below code taken from https://github.com/lh3/readfq/blob/master/readfq.py
'''

def readfq(fp): # this is a generator function
    last = None # this is a buffer keeping the last unprocessed line
    while True: # mimic closure; is it a bad idea?
        if not last: # the first record or a record following a fastq
            for l in fp: # search for the start of the next record
                if l[0] in '>@': # fasta/q header line
                    last = l[:-1] # save this line
                    break
        if not last: break
        name, seqs, last = last[1:], [], None
        for l in fp: # read the sequence
            if l[0] in '@+>':
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != '+': # this is a fasta record
            yield name, (''.join(seqs), None) # yield a fasta record
            if not last: break
        else: # this is a fastq record
            seq, leng, seqs = ''.join(seqs), 0, []
            for l in fp: # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq): # have read enough quality
                    last = None
                    yield name, (seq, ''.join(seqs)); # yield a fastq record
                    break
            if last: # reach EOF before reading enough quality
                yield name, (seq, None) # yield a fasta record instead
                break


def mkdir_p(path):
    try:
        os.makedirs(path)
        print("creating", path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def cigar_to_seq(cigar, query, ref):
    cigar_tuples = []
    result = re.split(r'[=DXSMI]+', cigar)
    i = 0
    for length in result[:-1]:
        i += len(length)
        type_ = cigar[i]
        i += 1
        cigar_tuples.append((int(length), type_ ))

    r_index = 0
    q_index = 0
    q_aln = []
    r_aln = []
    for length_ , type_ in cigar_tuples:
        if type_ == "=" or type_ == "X":
            q_aln.append(query[q_index : q_index + length_])
            r_aln.append(ref[r_index : r_index + length_])

            r_index += length_
            q_index += length_
        
        elif  type_ == "I":
            # insertion w.r.t. reference
            r_aln.append('-' * length_)
            q_aln.append(query[q_index: q_index + length_])
            #  only query index change
            q_index += length_

        elif type_ == 'D':
            # deletion w.r.t. reference
            r_aln.append(ref[r_index: r_index + length_])
            q_aln.append('-' * length_)
            #  only ref index change
            r_index += length_
        
        else:
            print("error")
            print(cigar)
            sys.exit()

    return  "".join([s for s in q_aln]), "".join([s for s in r_aln])










from __future__ import print_function
import os,sys
import argparse

import signal
from multiprocessing import Pool
import multiprocessing as mp

import operator
import functools
import errno
from time import time
from collections import deque
import sys
import itertools
import math

from modules import help_functions

D = {chr(i) : min( 10**( - (ord(chr(i)) - 33)/10.0 ), 0.79433)  for i in range(128)}
D_no_min = {chr(i) : 10**( - (ord(chr(i)) - 33)/10.0 )  for i in range(128)}

def expected_number_of_erroneous_kmers(quality_string, k):
    prob_error = [D[char_] for char_ in quality_string]
    window = deque([ (1.0 - p_e) for p_e in prob_error[:k]])
    # print(window)
    qurrent_prob_no_error = functools.reduce(operator.mul, window, 1)
    # print(qurrent_prob_no_error)
    sum_of_expectations = qurrent_prob_no_error # initialization 
    for p_e in prob_error[k:]:
        p_to_leave = window.popleft()
        # print(window)
        # print(p_to_leave, "!" in quality_string)
        qurrent_prob_no_error *= ((1.0 -p_e)/(p_to_leave))
        # print(qurrent_prob_no_error)
        sum_of_expectations += qurrent_prob_no_error
        window.append(1.0 -p_e)
    return len(quality_string) - k + 1 - sum_of_expectations 


def reverse_complement(string):
    rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'a':'t', 'c':'g', 'g':'c', 't':'a', 'N':'N', 'X':'X', 'n':'n', 'Y':'R', 'R':'Y', 'K':'M', 'M':'K', 'S':'S', 'W':'W', 'B':'V', 'V':'B', 'H':'D', 'D':'H', 'y':'r', 'r':'y', 'k':'m', 'm':'k', 's':'s', 'w':'w', 'b':'v', 'v':'b', 'h':'d', 'd':'h'}
    rev_comp = ''.join([rev_nuc[nucl] for nucl in reversed(string)])
    return(rev_comp)

def batch(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx + n, l)]


def calc_score_new(d):
    for key,value in d.items():
        l, k, q_threshold = value

    read_array = []
    error_rates = []
    for i, (acc, seq, qual) in enumerate(l):
        if i % 10000 == 0:
            print(i, "reads processed.")

        # skip very short reads or degenerate reads
        seq_hpol_comp = ''.join(ch for ch, _ in itertools.groupby(seq))
        if len(seq) < 2*k or len(seq_hpol_comp) < k:
            continue

        poisson_mean = sum([ qual.count(char_) * D_no_min[char_] for char_ in set(qual)])
        error_rate = poisson_mean/float(len(qual))
        if 10*-math.log(error_rate, 10) <= q_threshold:
            # print("Filtered read with:", 10*-math.log(error_rate, 10), error_rate)
            continue

        error_rates.append(error_rate)
        exp_errors_in_kmers = expected_number_of_erroneous_kmers(qual, k)
        p_no_error_in_kmers = 1.0 - exp_errors_in_kmers/ float((len(seq) - k +1))
        score =  p_no_error_in_kmers  * (len(seq) - k +1)
        read_array.append((acc, seq, qual, score) )
    return {key : (read_array, error_rates)}


def fastq_parallel(args):
    k = args.k
    q_threshold = args.quality_threshold
    error_rates = []
    reads = [ (acc,seq, qual) for acc, (seq, qual) in help_functions.readfq(open(args.fastq, 'r'))]
    start = time()
    read_chunk_size = int( len(reads)/args.nr_cores ) + 1
    read_batches = [b for b in batch(reads, read_chunk_size)]
    del reads
    ####### parallelize alignment #########
    # pool = Pool(processes=mp.cpu_count())
    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGINT, original_sigint_handler)
    mp.set_start_method('spawn')
    print(mp.get_context())
    print("Environment set:", mp.get_context())
    print("Using {0} cores.".format(args.nr_cores))
    start_multi = time()
    pool = Pool(processes=int(args.nr_cores))
    try:
        print([len(b) for b in read_batches])
        data = [ {i : (b,k, q_threshold)} for i, b in enumerate(read_batches)] #[ {i+1 :((cluster_batches[i], cluster_seq_origin_batches[i], read_batches[i], p_emp_probs, lowest_batch_index_db[i], i+1, args), {})} for i in range(len(read_batches))]
        res = pool.map_async(calc_score_new, data)
        score_results =res.get(999999999) # Without the timeout this blocking call ignores all signals.
    except KeyboardInterrupt:
        print("Caught KeyboardInterrupt, terminating workers")
        pool.terminate()
        sys.exit()
    else:
        pool.close()
    pool.join()

    print("Time elapesd multiprocessing:", time() - start_multi)
    read_array, error_rates = [], []

    for output_dict in score_results:
        for k, v in output_dict.items():
            r_a, err_rates = v
            print("Batch index", k)
            for item in r_a:
                read_array.append(item)
            for item2 in err_rates:
                error_rates.append(item2)

    read_array.sort(key=lambda x: x[3], reverse=True)
    error_rates.sort()
    return read_array, error_rates


def fastq_single_core(args):
    k = args.k
    q_threshold = args.quality_threshold
    error_rates = []
    read_array = []
    for i, (acc, (seq, qual)) in enumerate(help_functions.readfq(open(args.fastq, 'r'))):
        if i % 10000 == 0:
            print(i, "reads processed.")

        # skip very short reads or degenerate reads
        seq_hpol_comp = ''.join(ch for ch, _ in itertools.groupby(seq))
        if len(seq) < 2*k or len(seq_hpol_comp) < args.k:
            continue
        ########################
    
        exp_errors_in_kmers = expected_number_of_erroneous_kmers(qual, k)
        p_no_error_in_kmers = 1.0 - exp_errors_in_kmers/ float((len(seq) - k +1))
        score =  p_no_error_in_kmers  * (len(seq) - k +1)
        
        ## For (inferred) average error rate only, based on quality values
        ### These values are used in evaluations in the paper only, and are not used in clustering
        poisson_mean = sum([ qual.count(char_) * D_no_min[char_] for char_ in set(qual)])
        error_rate = poisson_mean/float(len(qual))
        if 10*-math.log(error_rate, 10) <= q_threshold:
            # print("Filtered read with:", 10*-math.log(error_rate, 10), error_rate)
            continue
        error_rates.append(error_rate)
        ##############################################
        
        read_array.append((acc, seq, qual, score) )

    read_array.sort(key=lambda x: x[3], reverse=True)
    return read_array, error_rates



def main(args):
    start = time()
    logfile = open(os.path.join(args.outfolder, "logfile.txt"), 'w')
    if os.path.isfile(args.outfile) and args.use_old_sorted_file:
        print("Using already existing sorted file in specified directory, in not intended, specify different outfolder or delete the current file.")
        return args.outfile

    elif args.fastq:
        if args.nr_cores > 1: 
            read_array, error_rates = fastq_parallel(args)
        else:
            read_array, error_rates = fastq_single_core(args)



    reads_sorted_outfile = open(args.outfile, "w")
    for i, (acc, seq, qual, score) in enumerate(read_array):
        reads_sorted_outfile.write("@{0}\n{1}\n+\n{2}\n".format(acc + "_{0}".format(score), seq, qual))
    reads_sorted_outfile.close()
    print(len(read_array), "reads passed quality critera (avg phred Q val over {0} and length > 2*k) and will be clustered.".format(args.quality_threshold))
    error_rates.sort()
    min_e = error_rates[0]
    max_e = error_rates[-1]
    median_e = error_rates[int(len(error_rates)/2)]
    mean_e = sum(error_rates)/len(error_rates)
    logfile.write("Lowest read error rate:{0}\n".format(min_e))
    logfile.write("Highest read error rate:{0}\n".format(max_e))
    logfile.write("Median read error rate:{0}\n".format(median_e))
    logfile.write("Mean read error rate:{0}\n".format(mean_e))
    logfile.write("\n")
    logfile.close()
    print("Sorted all reads in {0} seconds.".format(time() - start) )
    return reads_sorted_outfile.name


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Evaluate pacbio IsoSeq transcripts.")
    parser.add_argument('--fastq', type=str,  default=False, help='Path to consensus fastq file(s)')
    parser.add_argument('--outfile', type=str,  default=None, help='A fasta file with transcripts that are shared between samples and have perfect illumina support.')
    parser.add_argument('--k', type=int, default=15, help='kmer size')
    
    args = parser.parse_args()

    if (args.fastq and (args.flnc or args.ccs)):
        print("Either (1) only a fastq file, or (2) a ccs and a flnc file should be specified. ")
        sys.exit()

    if (args.flnc != False and args.ccs == False ) or (args.flnc == False and args.ccs != False ):
        print("qt-clust needs both the ccs.bam file produced by ccs and the flnc file produced by isoseq3 cluster. ")
        sys.exit()


    if len(sys.argv)==1:
        parser.print_help()
        sys.exit()
    path_, file_prefix = os.path.split(args.outfile)
    help_functions.mkdir_p(path_)

    main(args)


"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# using example setup file from https://github.com/pypa/sampleproject/blob/master/setup.py

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(

    name='NGSpeciesID',  # Required
    version='0.1.2',  # Required
    description='Reconstructs viral consensus sequences from a set of ONT reads.',  # Required
    long_description=long_description,  # Optional
    url='https://github.com/ksahlin/NGSpeciesID',  # Optional
    author='Kristoffer Sahlin',  # Optional
    author_email='ksahlin@math.su.se',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        #'Intended Audience :: Developers',
        #'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        #'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        #'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='viral sequeces ONT Oxford Nanopore Technologies long reads',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # If your package is for Python 2.7, and all versions of Python 3 starting with 3.4, write
    python_requires='!=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['parasail==1.1.10',
                      'edlib>=1.1.2'],  # Optional
    # dependency_links=[], # Optional
    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    # extras_require={  # Optional
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    # entry_points={  # Optional
    #     'console_scripts': [
    #         'NGSpeciesID=NGSpeciesID:main()',
    #     ],
    # },
    scripts=['NGSpeciesID'],
)




import edlib

from modules import help_functions

def reverse_complement(string):
    #rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N':'N', 'X':'X'}
    # Modified for Abyss output
    rev_nuc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'a':'t', 'c':'g', 'g':'c', 't':'a', 'N':'N', 'X':'X', 'n':'n', 'Y':'R', 'R':'Y', 'K':'M', 'M':'K', 'S':'S', 'W':'W', 'B':'V', 'V':'B', 'H':'D', 'D':'H', 'y':'r', 'r':'y', 'k':'m', 'm':'k', 's':'s', 'w':'w', 'b':'v', 'v':'b', 'h':'d', 'd':'h'}

    rev_comp = ''.join([rev_nuc[nucl] for nucl in reversed(string)])
    return(rev_comp)

def read_barcodes(primer_file):
    barcodes = { acc + '_fw' : seq.strip() for acc, (seq, _) in help_functions.readfq(open(primer_file, 'r'))}

    for acc, seq in list(barcodes.items()):
        print(acc, seq,acc[:-3])
        barcodes[acc[:-3] + '_rc'] = reverse_complement(seq.upper())

    print(barcodes)
    return barcodes

def get_universal_tails():
    barcodes = {'1_F_fw' : 'TTTCTGTTGGTGCTGATATTGC',
                 '2_R_rc' : 'ACTTGCCTGTCGCTCTATCTTC'}
    barcodes['1_F_rc'] = reverse_complement(barcodes['1_F_fw'])
    barcodes['2_R_fw'] = reverse_complement(barcodes['2_R_rc'])
    print(barcodes)
    return barcodes


def find_barcode_locations(center, barcodes, primer_max_ed):
    "Find barcodes in a center using edlib"
    
    # Creation of a IUPAC equivalence map for edlib to allow IUPAC code in primers
    # The IUPAC map was created with:
    # from Bio.Data import IUPACData
    # IUPAC_map = [(i, k) for i, j in IUPACData.ambiguous_dna_values.items() for k in j]
    IUPAC_map = [('A', 'A'), ('C', 'C'), ('G', 'G'), ('T', 'T'), ('M', 'A'), ('M', 'C'),
                 ('R', 'A'), ('R', 'G'), ('W', 'A'), ('W', 'T'), ('S', 'C'), ('S', 'G'),
                 ('Y', 'C'), ('Y', 'T'), ('K', 'G'), ('K', 'T'), ('V', 'A'), ('V', 'C'),
                 ('V', 'G'), ('H', 'A'), ('H', 'C'), ('H', 'T'), ('D', 'A'), ('D', 'G'),
                 ('D', 'T'), ('B', 'C'), ('B', 'G'), ('B', 'T'), ('X', 'G'), ('X', 'A'),
                 ('X', 'T'), ('X', 'C'), ('N', 'G'), ('N', 'A'), ('N', 'T'), ('N', 'C')]
    all_locations = []
    for primer_acc, primer_seq in barcodes.items():
        # print(primer_acc, primer_seq,center)
        # Add additionalEqualities=IUPAC_map allow edlib to understand IUPAC code
        result = edlib.align(primer_seq, center,
                             mode="HW", task="locations", k=primer_max_ed,
                             additionalEqualities=IUPAC_map)
        ed = result["editDistance"]
        locations = result["locations"]
        print(locations, ed)
        if locations:
            all_locations.append((primer_acc, locations[0][0], locations[0][1], ed))
    return all_locations


def remove_barcodes(centers, barcodes, args):
    """
        Modifies consensus sequences by copping of at barcode sites.
        This implies changing the datastructure centers with the modified consensus sequeces
    """
    
    for i, (nr_reads_in_cluster, c_id, center, reads_path_name) in enumerate(centers):

        # if consensus is smaller than 2*trim_window we set trim window to half the sequence
        if 2*args.trim_window > len(center):
            trim_window = len(center)//2
        else:
            trim_window = args.trim_window

        barcode_locations_beginning = find_barcode_locations(center[:trim_window], barcodes, args.primer_max_ed) 
        barcode_locations_end = find_barcode_locations(center[-trim_window:], barcodes, args.primer_max_ed) 
        print(center)
        
        cut_start = 0
        if barcode_locations_beginning:
            print("FOUND BARCODE BEGINNING", barcode_locations_beginning)
            for bc, start, stop, ed in barcode_locations_beginning:
                if stop > cut_start:
                    cut_start = stop
            
        cut_end = len(center)
        if barcode_locations_end:
            print("FOUND BARCODE END", barcode_locations_end)
            earliest_hit = len(center)
            for bc, start, stop, ed in barcode_locations_end:
                if start < earliest_hit:
                    earliest_hit = start
            cut_end = len(center) - (trim_window - earliest_hit)
        center = center[cut_start: cut_end]

        print(center, "NEW")
        print("cut start", cut_start, "cut end", cut_end)
        centers[i][2] = center

    ## Old code scanned all consensus and were prone to errors due to cutting directionality (befause of fwd or rev comp hits of the adapter)
    # for i, (nr_reads_in_cluster, c_id, center, reads_path_name) in enumerate(centers):
    #     barcode_locations = find_barcode_locations(center, barcodes, args.primer_max_ed) 
    #     if barcode_locations:
    #         print("FOUND BARCODE", barcode_locations)
    #         cut_start = 0
    #         cut_end = len(center)
    #         print(center)
    #         for bc, start, stop, ed in barcode_locations:
    #             # print(ed,bc, bc[-4], bc[-2:])
    #             if bc[-4] == 'F' and bc[-2:] == 'fw':
    #                 cut_start = stop
    #             elif bc[-4] == 'R' and bc[-2:] == 'fw': 
    #                 cut_end = start
    #             elif bc[-4] == 'R' and bc[-2:] == 'rc':
    #                 cut_start = stop
    #             elif bc[-4] == 'F' and bc[-2:] == 'rc': 
    #                 cut_end = start
    #             else:
    #                 print()
    #                 print("Primer file not in correct format!")
    #                 print()
    #         # print(center)
    #         center = center[cut_start: cut_end]
    #         print(center, "NEW")
    #         print("cut start", cut_start, "cut end", cut_end)
    #         centers[i][2] = center
